@@ -237,34 +237,19 @@ static inline int set_video_mode_Nala(struct pwc_device *pdev, int size, int fra
 	if (pEntry->alternate == 0)
 		return -EINVAL;
 
-	if (pEntry->compressed)
-		return -ENOENT; /* Not supported. */
-
 	memcpy(buf, pEntry->mode, 3);	
 	ret = send_video_command(pdev->udev, pdev->vendpoint, buf, 3);
 	if (ret < 0) {
-		Debug("Failed to send video command... %d\n", ret);
-		return ret;
+	      Debug("Failed to send video command... %d\n", ret);
+	      return ret;
+	}
+	if (pEntry->compressed && pdev->vpalette != VIDEO_PALETTE_RAW) {
+	      if (DEVICE_USE_CODEC1(pdev->type)) 
+		    pwc_dec1_init(pdev->type, pdev->release, buf, pdev->decompress_data);
+	      else
+		    pwc_dec23_init(pdev->type, pdev->release, buf, pdev->decompress_data);
 	}
-	if (pEntry->compressed && pdev->vpalette != VIDEO_PALETTE_RAW)
-	 {
-	   switch(pdev->type) {
-	     case 645:
-	     case 646:
-	       pwc_dec1_init(pdev->type, pdev->release, buf, pdev->decompress_data);
-	       break;
 
-	     case 675:
-	     case 680:
-	     case 690:
-	     case 720:
-	     case 730:
-	     case 740:
-	     case 750:
-	       pwc_dec23_init(pdev->type, pdev->release, buf, pdev->decompress_data);
-	       break;
-	   }
-	}
  
 	pdev->cmd_len = 3;
 	memcpy(pdev->cmd_buf, buf, 3);
@@ -445,25 +430,14 @@ int pwc_set_video_mode(struct pwc_device *pdev, int width, int height, int frame
 	}
 	Debug("decode_size = %d.\n", size);
 
-        ret = -EINVAL;
-	switch(pdev->type) {
-	case 645:
-	case 646:
+	if (DEVICE_USE_CODEC1(pdev->type)) {
 		ret = set_video_mode_Nala(pdev, size, frames);
-		break;
-
-	case 675:
-	case 680:
-	case 690:
-		ret = set_video_mode_Timon(pdev, size, frames, compression, snapshot);
-		break;
-	
-	case 720:
-	case 730:
-	case 740:
-	case 750:
+		
+	} else if (DEVICE_USE_CODEC3(pdev->type)) {
 		ret = set_video_mode_Kiara(pdev, size, frames, compression, snapshot);
-		break;
+
+	} else {
+		ret = set_video_mode_Timon(pdev, size, frames, compression, snapshot);
 	}
 	if (ret < 0) {
 		if (ret == -ENOENT)
@@ -628,7 +602,7 @@ int pwc_set_saturation(struct pwc_device *pdev, int value)
 
 /* AGC */
 
-static inline int pwc_set_agc(struct pwc_device *pdev, int mode, int value)
+int pwc_set_agc(struct pwc_device *pdev, int mode, int value)
 {
 	char buf;
 	int ret;
@@ -653,7 +627,7 @@ static inline int pwc_set_agc(struct pwc_device *pdev, int mode, int value)
 	return 0;
 }
 
-static inline int pwc_get_agc(struct pwc_device *pdev, int *value)
+int pwc_get_agc(struct pwc_device *pdev, int *value)
 {
 	unsigned char buf;
 	int ret;
@@ -683,7 +657,7 @@ static inline int pwc_get_agc(struct pwc_device *pdev, int *value)
 	return 0;
 }
 
-static inline int pwc_set_shutter_speed(struct pwc_device *pdev, int mode, int value)
+int pwc_set_shutter_speed(struct pwc_device *pdev, int mode, int value)
 {
 	char buf[2];
 	int speed, ret;
@@ -725,6 +699,33 @@ static inline int pwc_set_shutter_speed(struct pwc_device *pdev, int mode, int v
 	return ret;
 }	
 
+int pwc_get_shutter_speed(struct pwc_device *pdev, int *value)
+{
+	unsigned char buf[2];
+	int ret;
+	
+	ret = RecvControlMsg(GET_STATUS_CTL, READ_SHUTTER_FORMATTER, 2);
+	if (ret < 0)
+		return ret;
+	*value = buf[0] + (buf[1] << 8);
+	switch(pdev->type) {
+	case 675:
+	case 680:
+	case 690:
+		/* speed ranges from 0x0 to 0x290 (656) */
+		*value *= 100;
+		break;
+	case 720:
+	case 730:
+	case 740:
+	case 750:
+		/* speed seems to range from 0x0 to 0xff */
+		*value <<= 8;
+		break;
+	}
+	return 0;
+}
+
 
 /* POWER */
 
@@ -776,7 +777,7 @@ inline int pwc_restore_factory(struct pwc_device *pdev)
   * 03: manual
   * 04: auto
   */ 
-static inline int pwc_set_awb(struct pwc_device *pdev, int mode)
+int pwc_set_awb(struct pwc_device *pdev, int mode)
 {
 	char buf;
 	int ret;
@@ -796,7 +797,7 @@ static inline int pwc_set_awb(struct pwc_device *pdev, int mode)
 	return 0;
 }
 
-static inline int pwc_get_awb(struct pwc_device *pdev)
+int pwc_get_awb(struct pwc_device *pdev)
 {
 	unsigned char buf;
 	int ret;
@@ -1641,4 +1642,4 @@ int pwc_ioctl(struct pwc_device *pdev, unsigned int cmd, void *arg)
 }
 
 
-/* vim: set cinoptions={.5s,\:.5s,+.5s,t0,g0,^-2,e-2,n-2,p2s,(0,=.5s formatoptions=croql cindent shiftwidth=8 tabstop=8: */
+/* vim: set cinoptions= formatoptions=croql cindent shiftwidth=8 tabstop=8: */
