@@ -261,7 +261,7 @@ static void pwc_rvfree(void * mem, unsigned long size)
 
 static int pwc_allocate_buffers(struct pwc_device *pdev)
 {
-	int i;
+	int i, err;
 	void *kbuf;
 
 	Trace(TRACE_MEMORY, ">> pwc_allocate_buffers(pdev = 0x%p)\n", pdev);
@@ -310,31 +310,16 @@ static int pwc_allocate_buffers(struct pwc_device *pdev)
 	}
 	
 	/* Allocate decompressor table space */
-	kbuf = NULL;
-	switch (pdev->type)
-	 {
-	  case 675:
-	  case 680:
-	  case 690:
-	  case 720:
-	  case 730:
-	  case 740:
-	  case 750:
-	    Trace(TRACE_MEMORY,"private_data(%Zd)\n",sizeof(struct pwc_dec23_private));
-	    kbuf = kmalloc(sizeof(struct pwc_dec23_private), GFP_KERNEL);	/* Timon & Kiara */
-	    break;
-	  case 645:
-	  case 646:
-	    /* TODO & FIXME */
-	    kbuf = kmalloc(sizeof(struct pwc_dec23_private), GFP_KERNEL);
-	    break;
-	 }
-	if (kbuf == NULL) {
+	if (DEVICE_USE_CODEC1(pdev->type))
+		err = pwc_dec1_alloc(pdev);
+	else
+		err = pwc_dec23_alloc(pdev);
+
+	if (err) {
 	   Err("Failed to allocate decompress table.\n");
 	   return -ENOMEM;
 	}
-	pdev->decompress_data = kbuf;
-	
+
 	/* Allocate image buffer; double buffer for mmap() */
 	kbuf = pwc_rvmalloc(pwc_mbufs * PAGE_ALIGN(pdev->len_per_image));
 	if (kbuf == NULL) {
@@ -715,8 +700,10 @@ static void pwc_isoc_handler(struct urb *urb, struct pt_regs *regs)
 							pdev->vframes_error++;
 						}
 						if ((ptr[0] ^ pdev->vmirror) & 0x01) {
-							if (ptr[0] & 0x01)
+							if (ptr[0] & 0x01) {
+								pdev->snapshot_button_status = 1;
 								Info("Snapshot button pressed.\n");
+							}
 							else
 								Info("Snapshot button released.\n");
 						}
@@ -739,6 +726,17 @@ static void pwc_isoc_handler(struct urb *urb, struct pt_regs *regs)
 						if (fbuf->filled == 4)
 							pdev->drop_frames++;
 					}
+				       	else if (pdev->type == 740) {
+						unsigned char *ptr = (unsigned char *)fbuf->data;
+						if ((ptr[0] ^ pdev->vmirror) & 0x01) {
+							if (ptr[0] & 0x01) {
+								pdev->snapshot_button_status = 1;
+								Info("Snapshot button pressed.\n");
+							}
+							else
+								Info("Snapshot button released.\n");
+						}
+					}
 
 					/* In case we were instructed to drop the frame, do so silently.
 					   The buffer pointers are not updated either (but the counters are reset below).
@@ -1013,11 +1011,23 @@ static ssize_t store_pan_tilt(struct class_device *class_dev, const char *buf,
 static CLASS_DEVICE_ATTR(pan_tilt, S_IRUGO | S_IWUSR, show_pan_tilt,
 			 store_pan_tilt);
 
+static ssize_t show_snapshot_button_status(struct class_device *class_dev, char *buf)
+{
+	struct pwc_device *pdev = cd_to_pwc(class_dev);
+	int status = pdev->snapshot_button_status;
+	pdev->snapshot_button_status = 0;
+	return sprintf(buf, "%d\n", status);
+}
+
+static CLASS_DEVICE_ATTR(button, S_IRUGO | S_IWUSR, show_snapshot_button_status,
+			 NULL);
+
 static void pwc_create_sysfs_files(struct video_device *vdev)
 {
 	struct pwc_device *pdev = video_get_drvdata(vdev);
 	if (pdev->features & FEATURE_MOTOR_PANTILT)
 		video_device_create_file(vdev, &class_device_attr_pan_tilt);
+	video_device_create_file(vdev, &class_device_attr_button);
 }
 
 static void pwc_remove_sysfs_files(struct video_device *vdev)
@@ -1025,6 +1035,7 @@ static void pwc_remove_sysfs_files(struct video_device *vdev)
 	struct pwc_device *pdev = video_get_drvdata(vdev);
 	if (pdev->features & FEATURE_MOTOR_PANTILT)
 		video_device_remove_file(vdev, &class_device_attr_pan_tilt);
+	video_device_remove_file(vdev, &class_device_attr_button);
 }
 
 /***************************************************************************/
@@ -1166,22 +1177,10 @@ static int pwc_video_close(struct inode *inode, struct file *file)
 	if (pdev->vframe_count > 20)
 		Info("Closing video device: %d frames received, dumped %d frames, %d frames with errors.\n", pdev->vframe_count, pdev->vframes_dumped, pdev->vframes_error);
 
-	switch (pdev->type)
-	 {
-	  case 675:
-	  case 680:
-	  case 690:
-	  case 720:
-	  case 730:
-	  case 740:
-	  case 750:
-	    pwc_dec23_exit();	/* Timon & Kiara */
-	    break;
-	  case 645:
-	  case 646:
+	if (DEVICE_USE_CODEC1(pdev->type))
 	    pwc_dec1_exit();
-	    break;
-	 }
+	else
+	    pwc_dec23_exit();
 
 	pwc_isoc_cleanup(pdev);
 	pwc_free_buffers(pdev);
@@ -1267,7 +1266,7 @@ static ssize_t pwc_video_read(struct file *file, char __user *buf,
 
 	Trace(TRACE_READ, "Copying data to user space.\n");
 	if (pdev->vpalette == VIDEO_PALETTE_RAW)
-		bytes_to_read = pdev->frame_size;
+		bytes_to_read = pdev->frame_size + 4;
 	else
  		bytes_to_read = pdev->view.size;
 
