@@ -74,7 +74,6 @@ static struct v4l2_queryctrl pwc_controls[] = {
 	    .step    = 1,
 	    .default_value = 0,
 	},
-#if 0
         {
 	    .id      = V4L2_CID_RED_BALANCE,
 	    .type    = V4L2_CTRL_TYPE_INTEGER,
@@ -102,7 +101,33 @@ static struct v4l2_queryctrl pwc_controls[] = {
 	    .step    = 1,
 	    .default_value = 0,
 	},
-#endif
+	{
+	    .id      = V4L2_CID_EXPOSURE,
+	    .type    = V4L2_CTRL_TYPE_INTEGER,
+	    .name    = "Shutter Speed (Exposure)",
+	    .minimum = 0,
+	    .maximum = 256,
+	    .step    = 1,
+	    .default_value = 200,
+	},
+	{
+	    .id      = V4L2_CID_AUTOGAIN,
+	    .type    = V4L2_CTRL_TYPE_BOOLEAN,
+	    .name    = "Auto Gain Enabled",
+	    .minimum = 0,
+	    .maximum = 1,
+	    .step    = 1,
+	    .default_value = 1,
+	},
+	{
+	    .id      = V4L2_CID_GAIN,
+	    .type    = V4L2_CTRL_TYPE_INTEGER,
+	    .name    = "Gain Level",
+	    .minimum = 0,
+	    .maximum = 256,
+	    .step    = 1,
+	    .default_value = 0,
+	},
 #if XAWTV_HAS_BEEN_FIXED
         {
 	    .id      = V4L2_CID_PRIVATE_SAVE_USER,
@@ -142,6 +167,115 @@ static const char *v4l1_ioctls[] = {
         "SMICROCODE", "GVBIFMT", "SVBIFMT" };
 #define V4L1_IOCTLS ARRAY_SIZE(v4l1_ioctls)
 
+/* ioctl(VIDIOC_TRY_FMT) */
+static int pwc_vidioc_try_fmt(struct pwc_device *pdev, struct v4l2_format *f)
+{
+	Trace(TRACE_MODULE, "ioctl(VIDIOC_TRY_FMT)\n");
+
+	if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
+		Trace(TRACE_MODULE, "Bad video type must be V4L2_BUF_TYPE_VIDEO_CAPTURE\n");
+		return -EINVAL;
+	}
+
+	switch (f->fmt.pix.pixelformat) {
+		case V4L2_PIX_FMT_YUV420:
+			break;
+		case V4L2_PIX_FMT_PWC1:
+			if (DEVICE_USE_CODEC23(pdev->type)) {
+				Trace(TRACE_MODULE, "codec23 is only supported for new pwc webcam\n");
+				return -EINVAL;
+			}
+			break;
+		case V4L2_PIX_FMT_PWC2:
+			if (DEVICE_USE_CODEC1(pdev->type)) {
+				Trace(TRACE_MODULE, "codec1 is only supported for old pwc webcam\n");
+				return -EINVAL;
+			}
+			break;
+		default:
+			Trace(TRACE_MODULE, "Unsupported pixel format\n");
+			return -EINVAL;
+
+	}
+
+	if (f->fmt.pix.width > pdev->view_max.x)
+		f->fmt.pix.width = pdev->view_max.x;
+	else if (f->fmt.pix.width < pdev->view_min.x)
+		f->fmt.pix.width = pdev->view_min.x;
+
+	if (f->fmt.pix.height > pdev->view_max.y)
+		f->fmt.pix.height = pdev->view_max.y;
+	else if (f->fmt.pix.height < pdev->view_min.y)
+		f->fmt.pix.height = pdev->view_min.y;
+
+	return 0;
+}
+
+/* ioctl(VIDIOC_SET_FMT) */
+static int pwc_vidioc_set_fmt(struct pwc_device *pdev, struct v4l2_format *f)
+{
+	int ret, fps, snapshot, compression, pixelformat;
+
+	ret = pwc_vidioc_try_fmt(pdev, f);
+	if (ret<0)
+		return ret;
+
+	Trace(TRACE_MODULE, "ioctl(VIDIOC_S_FMT) \n");
+
+	pixelformat = f->fmt.pix.pixelformat;
+	compression = pdev->vcompression;
+	snapshot = 0;
+	fps = pdev->vframes;
+	if (f->fmt.pix.priv) {
+		compression = (f->fmt.pix.priv & PWC_QLT_MASK) >> PWC_QLT_SHIFT;
+		snapshot = f->fmt.pix.priv & PWC_FPS_SNAPSHOT;
+		fps = (f->fmt.pix.priv & PWC_FPS_FRMASK) >> PWC_FPS_SHIFT;
+		if (fps == 0)
+			fps = pdev->vframes;
+	}
+	snapshot = f->fmt.pix.priv & PWC_FPS_SNAPSHOT;
+
+	Trace(TRACE_MODULE,
+	      "Try to change format to: width=%d height=%d fps=%d "
+	      "compression=%d snapshot=%d format=%c%c%c%c\n",
+	      f->fmt.pix.width, f->fmt.pix.height, fps,
+	      compression, snapshot,
+	      (pixelformat>>24)&255,
+	      (pixelformat>>16)&255,
+	      (pixelformat>>8)&255,
+	      (pixelformat)&255);
+
+	ret = pwc_try_video_mode(pdev,
+			f->fmt.pix.width,
+			f->fmt.pix.height,
+			fps,
+			compression,
+			snapshot);
+
+	Trace(TRACE_MODULE, "pwc_try_video_mode(), return=%d\n", ret);
+
+	if (ret)
+		return ret;
+
+	memset(&f->fmt.pix,0,sizeof(struct v4l2_pix_format));
+	f->fmt.pix.width        = pdev->image.x;
+	f->fmt.pix.height       = pdev->image.y;
+	f->fmt.pix.field        = V4L2_FIELD_NONE;
+	f->fmt.pix.pixelformat  = pixelformat;
+	if (pixelformat == V4L2_PIX_FMT_YUV420) {
+		f->fmt.pix.bytesperline = (f->fmt.pix.width * 3)/2;
+		f->fmt.pix.sizeimage = f->fmt.pix.height * f->fmt.pix.bytesperline;
+		pdev->vpalette = VIDEO_PALETTE_YUV420P;
+	} else {
+		/* vbandlength contains 4 lines ...  */
+		f->fmt.pix.bytesperline = pdev->vbandlength/4;
+		f->fmt.pix.sizeimage = pdev->frame_size + sizeof(struct pwc_raw_frame);
+		pdev->vpalette = VIDEO_PALETTE_RAW;
+	}
+	return 0;
+
+}
+
 int pwc_video_do_ioctl(struct inode *inode, struct file *file,
 	 	       unsigned int cmd, void *arg)
 {
@@ -167,7 +301,7 @@ int pwc_video_do_ioctl(struct inode *inode, struct file *file,
                                vdev->minor, cmd,  v4l2_ioctl_names[_IOC_NR(cmd)]);
                         break;
                 default:
-                        printk(KERN_DEBUG PWC_NAME"%d: ioctl 0x%x (???)\n",
+                        printk(KERN_DEBUG PWC_NAME"%d: ioctl 0x%x (unknown)\n",
                                vdev->minor, cmd);
                 }
         }
@@ -563,6 +697,7 @@ int pwc_video_do_ioctl(struct inode *inode, struct file *file,
 		case VIDIOC_G_CTRL:
 		{
 			struct v4l2_control *c = arg;
+			int ret;
 
 			switch (c->id)
 			{
@@ -591,16 +726,15 @@ int pwc_video_do_ioctl(struct inode *inode, struct file *file,
 					if (c->value<0)
 						return -EINVAL;
 					return 0;
-#if 0
 				case V4L2_CID_RED_BALANCE:
-					ret = pwc_get_red_gain(pdev, c->value);
+					ret = pwc_get_red_gain(pdev, &c->value);
 					printk("pwc: V4L2_CID_RED_BALANCE()=%d\n",c->value);
 					if (ret<0)
 						return -EINVAL;
 					c->value >>= 8;
 					return 0;
 				case V4L2_CID_BLUE_BALANCE:
-					ret = pwc_get_blue_gain(pdev, c->value);
+					ret = pwc_get_blue_gain(pdev, &c->value);
 					printk("pwc: V4L2_CID_BLUE_BALANCE()=%d\n",c->value);
 					if (ret<0)
 						return -EINVAL;
@@ -613,7 +747,25 @@ int pwc_video_do_ioctl(struct inode *inode, struct file *file,
 						return -EINVAL;
 					c->value = (ret == PWC_WB_MANUAL)?0:1;
 					return 0;
-#endif
+				case V4L2_CID_GAIN:
+					ret = pwc_get_agc(pdev, &c->value);
+					printk("pwc: V4L2_CID_GAIN()=%d\n",c->value);
+					if (ret<0)
+						return -EINVAL;
+					c->value >>= 8;
+					return 0;
+				case V4L2_CID_AUTOGAIN:
+					ret = pwc_get_agc(pdev, &c->value);
+					if (ret<0)
+						return -EINVAL;
+					c->value = (c->value < 0)?1:0;
+					return 0;
+				case V4L2_CID_EXPOSURE:
+					ret = pwc_get_shutter_speed(pdev, &c->value);
+					if (ret<0)
+						return -EINVAL;
+					c->value >>= 8;
+					return 0;
 				case V4L2_CID_PRIVATE_SAVE_USER:
 				case V4L2_CID_PRIVATE_RESTORE_USER:
 				case V4L2_CID_PRIVATE_RESTORE_FACTORY:
@@ -623,22 +775,79 @@ int pwc_video_do_ioctl(struct inode *inode, struct file *file,
 		}
 		case VIDIOC_S_CTRL:
 		{
-			struct v4l2_queryctrl *c = arg;
+			struct v4l2_control *c = arg;
+			int ret;
 
 			switch (c->id)
 			{
+				case V4L2_CID_BRIGHTNESS:
+					c->value <<= 8;
+					ret = pwc_set_brightness(pdev, c->value);
+					if (ret<0)
+						return -EINVAL;
+					return 0;
+				case V4L2_CID_CONTRAST:
+					c->value <<= 8;
+					ret = pwc_set_contrast(pdev, c->value);
+					if (ret<0)
+						return -EINVAL;
+					return 0;
+				case V4L2_CID_GAMMA:
+					c->value <<= 8;
+					ret = pwc_set_gamma(pdev, c->value);
+					if (ret<0)
+						return -EINVAL;
+					return 0;
+				case V4L2_CID_RED_BALANCE:
+					c->value <<= 8;
+					ret = pwc_set_red_gain(pdev, c->value);
+					if (ret<0)
+						return -EINVAL;
+					return 0;
+				case V4L2_CID_BLUE_BALANCE:
+					c->value <<= 8;
+					ret = pwc_set_blue_gain(pdev, c->value);
+					if (ret<0)
+						return -EINVAL;
+					return 0;
+				case V4L2_CID_AUTO_WHITE_BALANCE:
+					c->value = (c->value == 0)?PWC_WB_MANUAL:PWC_WB_AUTO;
+					ret = pwc_set_awb(pdev, c->value);
+					if (ret<0)
+						return -EINVAL;
+					return 0;
+				case V4L2_CID_EXPOSURE:
+					c->value <<= 8;
+					ret = pwc_set_shutter_speed(pdev, c->value?0:1, c->value);
+					if (ret<0)
+						return -EINVAL;
+					return 0;
+				case V4L2_CID_AUTOGAIN:
+					/* autogain off means nothing without a gain */
+					if (c->value == 0)
+						return 0;
+					ret = pwc_set_agc(pdev, c->value, 0);
+					if (ret<0)
+						return -EINVAL;
+					return 0;
+				case V4L2_CID_GAIN:
+					c->value <<= 8;
+					ret = pwc_set_agc(pdev, 0, c->value);
+					if (ret<0)
+						return -EINVAL;
+					return 0;
 				case V4L2_CID_PRIVATE_SAVE_USER:
 					if (pwc_save_user(pdev))
-			                        return -EINVAL;
+						return -EINVAL;
 					return 0;
 				case V4L2_CID_PRIVATE_RESTORE_USER:
 					if (pwc_restore_user(pdev))
-			                        return -EINVAL;
+						return -EINVAL;
 					return 0;
 				case V4L2_CID_PRIVATE_RESTORE_FACTORY:
 					if (pwc_restore_factory(pdev))
-			                        return -EINVAL;
-					return 0;					
+						return -EINVAL;
+					return 0;
 			}
 			return -EINVAL;
 		}
@@ -646,20 +855,21 @@ int pwc_video_do_ioctl(struct inode *inode, struct file *file,
 		case VIDIOC_ENUM_FMT:
 		{
 		    	struct v4l2_fmtdesc *f = arg;
-			int i;
+			int index; 
 
 			if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
 			      return -EINVAL;
 
-			i = f->index;
 		    	/* We only support two format: the raw format, and YUV */
+			index = f->index;
 			memset(f,0,sizeof(struct v4l2_fmtdesc));
 			f->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
-			switch(i)
+			f->index = index;
+			switch(index)
 			{
 				case 0:
 					/* RAW format */
-			    		f->pixelformat = V4L2_PIX_FMT_PWC23;
+			    		f->pixelformat = pdev->type<=646?V4L2_PIX_FMT_PWC1:V4L2_PIX_FMT_PWC2;
 					f->flags = V4L2_FMT_FLAG_COMPRESSED;
 			    		strlcpy(f->description,"Raw Philips Webcam",sizeof(f->description));
 					break;
@@ -669,7 +879,7 @@ int pwc_video_do_ioctl(struct inode *inode, struct file *file,
 					break;
 			    	default:
 			      		return -EINVAL;
-			    }
+			}
 			return 0;
 		}
 
@@ -694,73 +904,10 @@ int pwc_video_do_ioctl(struct inode *inode, struct file *file,
 		}
 
 		case VIDIOC_TRY_FMT:
-		{
-			struct v4l2_format *f = arg;
-			Trace(TRACE_MODULE, "ioctl(VIDIOC_TRY_FMT)\n");
-			if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
-				return -EINVAL;
-
-			if (f->fmt.pix.pixelformat != V4L2_PIX_FMT_YUV420)
-				return -EINVAL;
-
-			if (f->fmt.pix.width > pdev->view_max.x)
-			       	f->fmt.pix.width = pdev->view_max.x;
-			else if (f->fmt.pix.width < pdev->view_min.x)
-				f->fmt.pix.width = pdev->view_min.x;
-			
-			if (f->fmt.pix.height > pdev->view_max.y)
-			       	f->fmt.pix.height = pdev->view_max.y;
-			else if (f->fmt.pix.height < pdev->view_min.y)
-			       	f->fmt.pix.height = pdev->view_min.y;
-			
-			return 0;
-		}
+			return pwc_vidioc_try_fmt(pdev, arg);
 
 	    	case VIDIOC_S_FMT:
-		{
-		    	struct v4l2_format *f = arg;
-			int ret, fps, snapshot, compression;
-
-			Trace(TRACE_MODULE, "ioctl(VIDIOC_S_FMT)\n");
-			if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
-			      return -EINVAL;
-
-			compression = (f->fmt.pix.priv & PWC_QLT_MASK) >> PWC_QLT_SHIFT;
-			snapshot = f->fmt.pix.priv & PWC_FPS_SNAPSHOT;
-			fps = (f->fmt.pix.priv & PWC_FPS_FRMASK) >> PWC_FPS_SHIFT;
-			if (fps == 0)
-				fps = pdev->vframes;
-			snapshot = f->fmt.pix.priv & PWC_FPS_SNAPSHOT;
-
-			Trace(TRACE_MODULE, "Try to change format to: width=%d height=%d fps=%d compression=%d snapshot=%d\n",
-					f->fmt.pix.width,
-					f->fmt.pix.height,
-					fps,
-					compression,
-					snapshot);
-
-			ret = pwc_try_video_mode(pdev,
-						 f->fmt.pix.width,
-						 f->fmt.pix.height,
-						 fps,
-						 compression,
-						 snapshot);
-
-			Trace(TRACE_MODULE, "leaving VIDIOC_S_FMT, return=%d\n",ret);
-			if (ret)
-			      return ret;
-
-			memset(&f->fmt.pix,0,sizeof(struct v4l2_pix_format));
-			f->fmt.pix.width        = pdev->image.x;
-			f->fmt.pix.height       = pdev->image.y;
-			f->fmt.pix.field        = V4L2_FIELD_NONE;
-			f->fmt.pix.pixelformat  = V4L2_PIX_FMT_YUV420;
-			f->fmt.pix.bytesperline =
-			    (f->fmt.pix.width * 3)/2;
-			f->fmt.pix.sizeimage =
-			    f->fmt.pix.height * f->fmt.pix.bytesperline;
-			return 0;
-		}
+			return pwc_vidioc_set_fmt(pdev, arg);
 
 		case VIDIOC_REQBUFS:
 		{
@@ -801,7 +948,10 @@ int pwc_video_do_ioctl(struct inode *inode, struct file *file,
 			buf->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
 			buf->index = index;
 			buf->m.offset = index * PAGE_ALIGN(pdev->len_per_image);
-			buf->bytesused = pdev->len_per_image;
+			if (pdev->vpalette == VIDEO_PALETTE_RAW)
+				buf->bytesused = pdev->frame_size + sizeof(struct pwc_raw_frame);
+			else
+				buf->bytesused = pdev->len_per_image;
 			buf->field = V4L2_FIELD_NONE;
 			buf->memory = V4L2_MEMORY_MMAP;
 			//buf->flags = V4L2_BUF_FLAG_MAPPED;
@@ -875,7 +1025,10 @@ int pwc_video_do_ioctl(struct inode *inode, struct file *file,
 			Trace(TRACE_READ, "VIDIOC_DQBUF: after pwc_handle_frame\n");
 
 			buf->index = pdev->fill_image;
-			buf->bytesused = pdev->vpalette == VIDEO_PALETTE_RAW ? pdev->frame_size: pdev->len_per_image;
+			if (pdev->vpalette == VIDEO_PALETTE_RAW)
+				buf->bytesused = pdev->frame_size + sizeof(struct pwc_raw_frame);
+			else
+				buf->bytesused = pdev->len_per_image;
 			buf->flags = V4L2_BUF_FLAG_MAPPED;
 			buf->field = V4L2_FIELD_NONE;
 			do_gettimeofday(&buf->timestamp);
